//Jeannie Albrecht
//Reads in a request for resources
//Returns a list of machines that satisfy the request

//package bamboo.rd;
import java.util.*;
import org.jdom.*;
import java.io.*;

public class FindHosts {
    Vector nodeList=null, groupList=null, stack=null;
    HashMap sites=null, groupToNodes=null, latencies=null, 
	grouphash=null, replyStats=null, masterGrouphash=null;
    Document doc=null; 
    List groups=null, constraintList=null;    
    boolean DEBUG = false, DEBUG2 = false;
    String result="";
    double groupcounter=0.0, groupcounter2=0.0, lowestcost=99999.0, matchcounter=0.0;
    Stack constraintStack=null;
    long starttime;

    public FindHosts(Document requestdoc, long starttime) {
	constraintStack=new Stack();
        nodeList=new Vector();
        groupList=new Vector();
        grouphash=new HashMap();
        stack=new Vector();
        doc=requestdoc;
	constraintList=doc.getRootElement().getChildren("constraint");  
        groups = doc.getRootElement().getChildren("group");     
	masterGrouphash=new HashMap();
	this.starttime=starttime;
    }

    public void setDebug2(boolean f) {
	DEBUG2=f;
    }

    public void setDebug(boolean f) {
	DEBUG=f;
    }

    public Vector getMasterList(HashMap replyStats, HashMap groupToNodes, HashMap latencies) {
	//Returns a Vector of Vectors, where each group gets a vector that corresponds
     	// to the master list of nodes (of type Node)
	Iterator lit=null, it=null;
	Element e=null;
	Vector v=null;
	String groupname="";	
	HashMap hm=null;

	if(this.replyStats==null) {
	    this.replyStats=replyStats;
	}

	if(this.groupToNodes==null) {
	    this.groupToNodes=groupToNodes;
	}

	if(this.latencies==null) {
	    this.latencies=latencies;	   
	}

     	lit=groups.iterator();

     	while(lit.hasNext()) {
     	    //loop through group requests
     	    e=(Element)lit.next();
	    
     	    //Now find hashmap for first group
	    v=new Vector();
     	    groupname=e.getChildTextTrim("name");
     	    hm=(HashMap)groupToNodes.get(groupname);
     	    it = hm.keySet().iterator();
     	    while (it.hasNext()) {
     		String n = (String)it.next();
		double c=((Double)hm.get(n)).doubleValue();
		Node node=new Node(n,c);
		v.add(node);
	    }

	    //sort v by cost
	    QuickSort qs=new QuickSort("Node");
	    v=qs.sort(v);
	    
	    nodeList.add(v);
     	}	
    
	if (DEBUG) System.err.println("getMasterList before calling getPingData nodeList");
	return nodeList;
    }            
        
    public Vector createGroups(Vector allNodes, long time_limit, int totalnumgroups) {
	time_limit=time_limit*1000;
	
	Iterator i=null;
	boolean done=false, stop=false;
	long curr_time=System.currentTimeMillis();
	int groupnum=0;
	int num_groups=allNodes.size();
	boolean ignorelatency=false;
	
	if(DEBUG)System.err.println("Current time:  "+curr_time);
	if(DEBUG)System.err.println("Time to bail: "+(curr_time+time_limit));
	
	i = groups.iterator();
	int num_searches=0;
	groupnum=0;

	//loop through group definitions in xml and find groups that satisfy requirements
	//loop for each group in request
	while (i.hasNext()) {
	    Vector v=new Vector();
	    Vector names=new Vector();
	    Element g = (Element)i.next(); 
	    String name=g.getChildTextTrim("name");
	    int nummac= 999;
	    if(!(g.getChildTextTrim("num_machines")).equals("all")) {
		nummac = (int)(Double.valueOf(g.getChildTextTrim("num_machines")).doubleValue());
	    }
	    else {
		result+="You have selected \"all\" machines.  \n";
		result+="Inter-node and cross group constraints are not supported!\n"; 
	    }

	    String temp=g.getChildTextTrim("latency");
	    if(temp==null || temp.indexOf(",")<0 || nummac==999) {
		temp=" 0.0,0.0,999999.0,999999.0,0.0";
		ignorelatency=true;
	    }

	    StringTokenizer st=new StringTokenizer(temp, ",");
	    double abs_min=Double.valueOf(st.nextToken()).doubleValue();
	    double ideal_min=Double.valueOf(st.nextToken()).doubleValue();
	    double ideal_max=Double.valueOf(st.nextToken()).doubleValue();
	    double abs_max=Double.valueOf(st.nextToken()).doubleValue();
	    double lat_cost=Double.valueOf(st.nextToken()).doubleValue();

	    Group g1=null;
	    Vector tempList=null, vec=null;
	    
	    //create vector of potential machine names for next group
	    for (int q=0; q<((Vector)allNodes.elementAt(groupnum)).size(); q++) {
		names.add(((Node)((Vector)allNodes.elementAt(groupnum)).elementAt(q)).getName());
	    }
	    
	    if(names.size()==0) {
		stop=true;
	    }
	    
	    //start searching for groups
	    while(!stop) {
		if (DEBUG) System.err.println("First node in candidate list is  "+names.elementAt(0));
		g1 = new Group (name, nummac, abs_min, abs_max);
		
		tempList=new Vector();			      			      
		vec = recCheckLat(names, 0, tempList, nummac, abs_min, abs_max, ignorelatency);
		
		//if we find a good group
		if(vec.size()==nummac) {
		    //compute cost
		    double cost=computeCost(vec, ideal_min, ideal_max, abs_min, abs_max, lat_cost, name, ignorelatency);
		    Vector tempvec = new Vector();
		    for (int j=0; j<vec.size(); j++) {
			tempvec.add((String)vec.elementAt(j));
		    }
		    g1.setList(tempvec);
		    g1.setCost(cost);
		    g1.setID(num_searches);
		    masterGrouphash.put(new Integer(num_searches), g1);
		    num_searches++;
		    v.add(g1);

		    //Continue searching where we left off
		    String n=(String)(g1.getList()).elementAt(0);
		    while(!(names.elementAt(0)).equals(n)) {
			names.remove(0);
		    }
		    names.remove(0);
		    vec=null;
		}
		    
		else { 			      
		    //didn't find any groups!!  this is bad.  we are done.
		    if (DEBUG) System.err.println("Group found had size "+vec.size()+" but not "+nummac);

		    if(vec.size()>0 && nummac==999) {
			g1.setMachines(vec.size());
			
			//compute cost but ALWAYS ignore latency
			double cost=computeCost(vec, ideal_min, ideal_max, abs_min, abs_max, lat_cost, name, true);
			Vector tempvec = new Vector();
			for (int j=0; j<vec.size(); j++) {
			    tempvec.add((String)vec.elementAt(j));
			}
			g1.setList(tempvec);
			g1.setCost(cost);
			g1.setID(num_searches);
			masterGrouphash.put(new Integer(num_searches), g1);
			num_searches++;
			v.add(g1);
			
			vec=null;
		    }
		    stop=true;
		}			   		      			      			  
		
		if (System.currentTimeMillis()>(curr_time+(groupnum+1)*(time_limit/(num_groups*2)))) {
		    if(DEBUG) System.err.println("We are out of time.");
		    stop=true;
		}
		
		else {
		    if(DEBUG) System.err.println("Curr time: "+System.currentTimeMillis());
		    if(DEBUG) System.err.println("Time elapsed is "
						 +(System.currentTimeMillis()-curr_time)+"  which is "+
						 "less than "+(groupnum+1)*(time_limit/(num_groups/2))+" so we are continuing.");
		}
		
		if (names.size()==0) {
		    stop=true;
		}
	    }
	    
	    if(v.size()>0) {
		//we found some groups that worked
		QuickSort qs = new QuickSort("Group");
		v=qs.sort(v);
		grouphash.put(name,v);
	    }
	    
	    else {
		result+="Unable to find enough nodes for "+name+"!\n\n";
		result+="Request cannot be satisfied! \n";
		System.err.println(result);
		grouphash.put(name,null);
		groupList.removeAllElements();
		return groupList;
	    }
	    
	    //move on to next group.  reset variables
	    tempList=null;   vec=null;  names=null;
	    g1=null;   g=null;   v=null;
	    stop=false;		      
	    groupnum++;
	    if (DEBUG) System.err.println();
	}
	
	//print them
	if (DEBUG2) {
	    System.err.println("Done finding groups.");
	    Iterator keyIt = grouphash.keySet().iterator();
	    while(keyIt.hasNext()) {
		String next = (String)keyIt.next();
		Vector v=(Vector)grouphash.get(next);  
		System.err.println("\n"+next);
		for (int j=0; j<v.size();j++) {
		    System.err.println(j+" ID: "+((Group)v.elementAt(j)).getID()+" Cost: "+((Group)v.elementAt(j)).getCost());
		    for (int jj=0; jj<((Group)v.elementAt(j)).getList().size(); jj++) {
			System.err.print(" "+((Group)v.elementAt(j)).getList().elementAt(jj));
		    }
		    System.err.println();
		}
	    }
	}

	//now we have to check for any cross group constraints
	if(grouphash.size()!=0) {
	    Vector v = new Vector();
	    Iterator hashIt = grouphash.keySet().iterator();
	    while(hashIt.hasNext()) {
		v.add(grouphash.get(hashIt.next()));
	    }
	    //done=checkConstraints(v, new Vector());
	    done=checkConstraints(v, curr_time+time_limit, totalnumgroups);
	}
	
	else {
	    if (DEBUG) System.err.println("Request cannot be satisfied.");
	}
	
	if(groupList.size()==0 || !done) {
	    if (DEBUG) System.err.println("Cross group constraints failed.");
	    
	    if(System.currentTimeMillis()>(curr_time+time_limit)) {
		result+="Cross group constraint was not met, and time has expired.";
	    }
	    
	    result+="Request cannot be satisfied! \n";
	}
	
	//if(done) {
	//  finalVec=new Vector();
	// 	    //	    for (int j=0; j<groupList.size(); j++) {
	// 	    while(groupList.size()>0) {
	// 		matchcounter++;
	// 		//		Vector g = (Vector)groupList.elementAt(j);
	// 		Vector g = (Vector)groupList.remove(0);
	// 		Vector tempvec = new Vector();
	// 		tempvec.removeAllElements();
	// 		for (int q=0; q<g.size(); q++) {
	// 		    tempvec.add((Group)masterGrouphash.get((String)g.elementAt(q)));
	// 		}
	// 		double matchcost=checkLatencyConstraints(tempvec);
	// 		if(matchcost!=-1) {		
	// 		    //Vector parentVec=(Vector)groupList.elementAt(j);
	// 		    Vector parentVec=tempvec;
	// 		    double tempcost=matchcost;
	// 		    if (DEBUG2) System.err.print(matchcounter);
	// 		    for (int s=0; s<parentVec.size(); s++) {
	// 			tempcost+=((Group)parentVec.elementAt(s)).getCost();
	// 			if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
	// 					 +((Group)parentVec.elementAt(s)).getID());
	// 		    }
	// 		    if(DEBUG2) System.err.print(" "+tempcost);
	
	// 		    if(tempcost<lowestcost) {
	// 			lowestcost=tempcost;
	// 			finalVec.insertElementAt(parentVec,0);
	// 			if(DEBUG2) System.err.println(" ***new lowest cost found "+lowestcost);
	// 		    }
	// 		    else {
	// 			finalVec.add(parentVec);
	// 			if(DEBUG2) System.err.println();
	// 		    }
	// 		}
	// 	    }
	//}
	
	allNodes=null;
	return groupList;
    }
    
    public Vector createGroups2 (Vector nodes, long time_limit, int totalnumgroups) {	
 	// time_limit=time_limit*1000;
	
//  	Iterator i=null, latit=null;
// 	Iterator it = latencies.keySet().iterator();	
// 	long curr_time = System.currentTimeMillis();	
// 	int numhosts=0;
// 	int num_groups = nodes.size();
// 	int groupnum=0;
// 	boolean ignorelatency=false, done=false, stop=false;
// 	Vector lats = new Vector();
// 	Vector names = null;
// 	Vector v2 = null;
// 	Vector allNodes=new Vector();		
// 	Element e=null;

// 	if(DEBUG)System.err.println("Current time:  "+curr_time);
// 	if(DEBUG)System.err.println("Time to bail: "+(curr_time+time_limit));	

// 	//sort link latencies
// 	while(it.hasNext()) {
// 	    String key = (String)it.next();
// 	    double lat = ((Double)latencies.get(key)).doubleValue();
// 	    StringTokenizer st = new StringTokenizer(key, "s:d");
// 	    String src = st.nextToken();
// 	    String dst = st.nextToken();

// 	    if(!src.equals(dst)) {
// 		Link l = new Link(src, dst, lat);
// 		lats.add(l);	    
// 	    }
// 	}

// 	QuickSort qs=new QuickSort("Link");
// 	lats=qs.sort(lats); 	
		
// 	//create vector of best endpoints based on lowest latency
// 	Vector lats2 = new Vector();
// 	latit = lats.iterator();
// 	while(latit.hasNext()) {
// 	    Link l = (Link)latit.next();
// 	    if(!lats2.contains(l.getSrc())) 
// 		lats2.add(l.getSrc());

// 	    if(!lats2.contains(l.getDst()))
// 		lats2.add(l.getDst());
// 	}

// 	i=groups.iterator();
// 	int counter=0;
// 	int latcounter=0;

// 	Vector nodes2 = new Vector();
// 	//reshift names vector to favor low latency
// 	while (i.hasNext()) {	    
// 	    i.next();
// 	    names = new Vector();

// 	    //create vector of potential machine names for next group
// 	    for (int q=0; q<((Vector)nodes.elementAt(counter)).size(); q++) {
// 		names.add(((Node)((Vector)nodes.elementAt(counter)).elementAt(q)).getName());		
// 	    }
	    
// 	    done=false;
// 	    while(!done) {
// 		String curname = (String)lats2.elementAt(latcounter);
// 		if(names.contains(curname)) {
// 		    names.remove(curname);
// 		    names.insertElementAt(curname, 0);
// 		    lats2.remove(curname);
// 		    done=true;
// 		}
// 		latcounter++;
// 	    }
// 	    nodes2.add(names);
// 	    counter++;
// 	}
// 	done=false;

// 	allNodes=nodes2;

// 	i = groups.iterator();
// 	int num_searches=0;
	
// 	//loop through group definitions in xml and find groups that satisfy requirements
// 	//loop for each group in request
// 	while (i.hasNext()) {
// 	    Vector v=new Vector();
// 	    names=new Vector();
// 	    Element g = (Element)i.next(); 
// 	    String name=g.getChildTextTrim("name");
// 	    int nummac= (int)(Double.valueOf(g.getChildTextTrim("num_machines")).doubleValue());
	    
// 	    String temp=g.getChildTextTrim("latency");
// 	    if(temp==null || temp.indexOf(",")<0) {
// 		temp=" 0.0,0.0,999999.0,999999.0,0.0";
// 		ignorelatency=true;
// 	    }
	    
// 	    StringTokenizer st=new StringTokenizer(temp, ",");
// 	    double abs_min=Double.valueOf(st.nextToken()).doubleValue();
// 	    double ideal_min=Double.valueOf(st.nextToken()).doubleValue();
// 	    double ideal_max=Double.valueOf(st.nextToken()).doubleValue();
// 	    double abs_max=Double.valueOf(st.nextToken()).doubleValue();
// 	    double lat_cost=Double.valueOf(st.nextToken()).doubleValue();

// 	    Group g1=null;
// 	    Vector tempList=null, vec=null;
	    
// 	    //create vector of potential machine names for next group
// 	    for (int q=0; q<((Vector)allNodes.elementAt(groupnum)).size(); q++) {
// 		names.add(((Node)((Vector)allNodes.elementAt(groupnum)).elementAt(q)).getName());
// 	    }
	    
// 	    if(names.size()==0) {
// 		stop=true;
// 	    }
	    
// 	    //start searching for groups
// 	    while(!stop) {
// 		if (DEBUG) System.err.println("First node in candidate list is  "+names.elementAt(0));
// 		g1 = new Group (name, nummac, abs_min, abs_max);
		
// 		tempList=new Vector();			      			      
// 		vec = recCheckLat(names, 0, tempList, nummac, abs_min, abs_max, ignorelatency);
		
// 		//if we find a good group
// 		if(vec.size()==nummac) {
// 		    //compute cost
// 		    double cost=computeCost(vec, ideal_min, ideal_max, abs_min, abs_max, lat_cost, name, ignorelatency);
// 		    Vector tempvec = new Vector();
// 		    for (int j=0; j<vec.size(); j++) {
// 			tempvec.add((String)vec.elementAt(j));
// 		    }
// 		    g1.setList(tempvec);
// 		    g1.setCost(cost);
// 		    g1.setID(num_searches);
// 		    masterGrouphash.put(new Integer(num_searches), g1);		
// 		    num_searches++;		    
// 		    v.add(g1);
		    
// 		    //Continue searching where we left off
// 		    String n=(String)(g1.getList()).elementAt(0);
// 		    while(!(names.elementAt(0)).equals(n)) {
// 			names.remove(0);
// 		    }
// 		    names.remove(0);
// 		    vec=null;
// 		}
				
// 		else { 			      
// 		    //didn't find any groups!!  this is bad.  we are done.
// 		    if (DEBUG) System.err.println("Group found had size "+vec.size()+" but not "+nummac);  
// 		    stop=true;
// 		}			   		      			      			  
		
// 		if (System.currentTimeMillis()>(curr_time+(groupnum+1)*(time_limit/num_groups))) {
// 		    if(DEBUG) System.err.println("We are out of time.");
// 		    stop=true;
// 		}
		
// 		else {
// 		    if(DEBUG) System.err.println("Curr time: "+System.currentTimeMillis());
// 		    if(DEBUG) System.err.println("Time elapsed is "
// 						 +(System.currentTimeMillis()-curr_time)+"  which is "+
// 						 "less than "+(groupnum+1)*(time_limit/num_groups)+" so we are continuing.");
// 		}
		
// 		if (names.size()==0) {
// 		    stop=true;
// 		}
// 	    }
	    
// 	    if(v.size()>0) {
// 		//we found some groups that worked
// 		grouphash.put(name,v);
// 	    }
	    
// 	    else {
// 		result+="Unable to find enough nodes for "+name+"!\n\n";
// 		result+="Request cannot be satisfied! \n";

// 		grouphash.put(name,null);
// 		groupList.removeAllElements();
// 		return groupList;
// 	    }
	    
// 	    //move on to next group.  reset variables
// 	    tempList=null;   vec=null;  names=null;
// 	    g1=null;   g=null;   v=null;
// 	    stop=false;		      
// 	    groupnum++;
// 	    if (DEBUG) System.err.println();
// 	}
	
// 	//print them
// 	if (DEBUG2) {
// 	    System.err.println("Done finding groups.");
// 	    Iterator keyIt = grouphash.keySet().iterator();
// 	    while(keyIt.hasNext()) {
// 		String next = (String)keyIt.next();
// 		Vector v=(Vector)grouphash.get(next);  
// 		System.err.println("\n"+next);
// 		for (int j=0; j<v.size();j++) {
// 		    System.err.println(j+" ID: "+((Group)v.elementAt(j)).getID()+" Cost: "+((Group)v.elementAt(j)).getCost());
// 		    for (int jj=0; jj<((Group)v.elementAt(j)).getList().size(); jj++) {
// 			System.err.print(" "+((Group)v.elementAt(j)).getList().elementAt(jj));
// 		    }
// 		    System.err.println();
// 		}
// 	    }
// 	}

// 	//now we have to check for any cross group constraints
// 	if(grouphash.size()!=0) {
// 	    Vector v = new Vector();
// 	    Iterator hashIt = grouphash.keySet().iterator();
// 	    while(hashIt.hasNext()) {
// 		v.add(grouphash.get(hashIt.next()));
// 	    }
// 	    //done=checkConstraints(v, new Vector());
// 	    done=checkConstraints(v);
// 	}
	
// 	else {
// 	    if (DEBUG) System.err.println("Request cannot be satisfied.");
// 	}
	
// 	if(groupList.size()==0 || !done) {
// 	    if (DEBUG) System.err.println("Cross group constraints failed.");
	    
// 	    if(System.currentTimeMillis()>(curr_time+time_limit)) {
// 		result+="Cross group constraint was not met, and time has expired.";
// 	    }
	    
// 	    result+="Request cannot be satisfied! \n";
// 	}
	
// 	allNodes=null;
// 	return groupList;
//     }

//     public Vector checkNumHosts(Vector nodes) {
// 	Element e=null;
// 	Vector allNodes=new Vector(), v2=null;
// 	int numhosts=0, counter=0;
	
// 	Iterator i=groups.iterator();
	
//      	while(i.hasNext()) {
//      	    //loop through group requests
// 	    sites=new HashMap();
//      	    e=(Element)i.next();
// 	    v2=new Vector();

// 	    //find numhosts	
// 	    numhosts = Integer.parseInt(e.getChildTextTrim("numhosts"));
// 	    if (DEBUG) System.err.println(e.getChildTextTrim("name")+" numhosts="+numhosts);	    

// 	    Vector nodes2 = (Vector)nodes.elementAt(counter);
// 	    Iterator vit=nodes2.iterator();
// 	    while(vit.hasNext()) {
// 		Node node=(Node)vit.next();
// 		String temp=node.getName();
// 		HashMap h=(HashMap)replyStats.get(temp);			
// 		Double svsite = (Double)h.get("site");
// 		double site = svsite.doubleValue();
// 		Integer count=((Integer)sites.get(new Double(site))); 
// 		int ccount=0;
	    		
// 		//check numhosts
// 		if(count==null) {
// 		    sites.put(new Double(site),new Integer(1));
// 		}
// 		else { 		
// 		    ccount=count.intValue();
// 		    ccount++;
// 		    count=new Integer(ccount);
// 		    sites.put(new Double(site), count);
// 		}
		
// 		//if we have seen too many from this site, remove node from candidate set
// 		//otherwise, add to v2
// 		if(ccount<=numhosts) {
// 		    v2.add(node);
// 		}
// 		else {
// 		    if(DEBUG)System.err.println("Skipping "+temp+" because we have seen too many from that site");
// 		}
// 	    }
// 	    allNodes.add(v2);
// 	    v2=null;
// 	    sites=null;
// 	    counter++;
// 	}
// 	return allNodes;
	return null;
    }
    
    public double computeCost(Vector vec, double ideal_min, double ideal_max, double abs_min, double abs_max,
			      double lat_cost, String name, boolean ignorelatency) {
	double cost=0.0, lat=0.0;
	int numlinks=0;
	
	if(!ignorelatency) {
	    //first check latency cost
	    for (int a=0; a<vec.size()-1; a++) {
		for (int b=a+1; b<vec.size(); b++) {
		    String src_name=(String)vec.elementAt(a);
		    String dest_name=(String)vec.elementAt(b);		
		    
		    if(latencies.containsKey("s:"+src_name+"d:"+dest_name) && 
		       latencies.containsKey("s:"+dest_name+"d:"+src_name)&& (a!=b)) {
			//find avg latency for src->dst and dst->src
			lat=((Double)latencies.get("s:"+src_name+"d:"
						   +dest_name)).doubleValue();
			lat+=((Double)latencies.get("s:"+dest_name+"d:"+src_name)).doubleValue();
			lat=lat/2.0;
			numlinks++;
		    }
		    else if (latencies.containsKey("s:"+dest_name+"d:"+src_name)&& (a!=b)) {
			lat=((Double)latencies.get("s:"+dest_name+"d:"+src_name)).doubleValue();
			numlinks++;
		    }
		    else if (latencies.containsKey("s:"+src_name+"d:"+dest_name)&& (a!=b)) {
			lat=((Double)latencies.get("s:"+src_name+"d:"+dest_name)).doubleValue();
			numlinks++;
		    }
		       		    
		    else {  //really bad news
			if (DEBUG) System.err.println("Missing link info for "+src_name+"-> "+dest_name);
		    }

		    double denom=0;
		    if(lat>abs_min && lat<abs_max) {
			if(lat<ideal_min) {				    
			    denom=ideal_min-abs_min;
			    if(denom!=0)
				cost+=((ideal_min-lat)/denom)*lat_cost;
			}
			else  if(lat>ideal_max) {
			    denom=abs_max-ideal_max;
			    if(denom!=0)
				cost+=((lat-ideal_max)/denom)*lat_cost;
			}
		    }
		}
	    }
	}

	//now get other costs
	for (int a=0; a<vec.size();a++) {
	    HashMap hm=(HashMap)groupToNodes.get(name);
	    String n=(String)vec.elementAt(a);
	    double c=((Double)hm.get(n)).doubleValue();
	    if (DEBUG) System.err.println("Cost is "+c+" + "+cost+"="+(cost+c)+" numlinks="+numlinks);
	    cost+=c;
	    cost=((int)(cost*100))/100.0;
	}				   				  	
	return cost;
    }

    public Vector recCheckLat(Vector listNodes, int count, Vector newVec, int max, 
			      double min_latency, double max_latency, boolean ignorelatency) {		      
	if (DEBUG) {
	    System.err.print("\nlistNodes:");
	    for (int i=0; i<listNodes.size(); i++) {
		System.err.print(listNodes.elementAt(i)+"   ");
	    }
	    System.err.println();  
	    System.err.print("New Vec:");
	    for (int i=0; i<newVec.size(); i++) {
		System.err.print(newVec.elementAt(i)+"   ");
	    }
	    System.err.println();
	}
	
	Vector newList=new Vector();
	Vector oldList = new Vector();
	for (int i=0; i<listNodes.size(); i++) {
	    oldList.add(listNodes.elementAt(i));
	}
	
	//Pop first element off list of potential nodes 
	String first=(String)listNodes.elementAt(0);
	oldList.removeElementAt(0);

	//Find links in oldList that meet latency requirement for node in question
	int ii=0;
	for (ii=0; ii<oldList.size(); ii++) {
	    String repname=(String)oldList.elementAt(ii);
	    String key="s:"+first+"d:"+repname;
	    String key2="s:"+repname+"d:"+first;
	    
	    if(ignorelatency) {
		newList.add((String)oldList.elementAt(ii));
	    }
	    
	    else if(latencies.containsKey(key)) {
		//if (DEBUG) System.err.println("Found "+key);
		double lat=((Double)latencies.get(key)).doubleValue();
		if (lat < max_latency && lat > min_latency) {
		    //    key="s:"+repname+"d:"+first;
		    //if(latencies.containsKey(key)) {
		    //lat=((Double)latencies.get(key)).doubleValue();
		    //if (lat < max_latency && lat > min_latency) {
		    newList.add((String)oldList.elementAt(ii));
		    //}
		    //}
		}
	    }
	    
	    //adding because latency measurements are bad!	    
	    else if(latencies.containsKey(key2)) {
		key="s:"+repname+"d:"+first;
		//if (DEBUG) System.err.println("Found "+key);
		double lat=((Double)latencies.get(key)).doubleValue();
		if (lat < max_latency && lat > min_latency) {
		    //key="s:"+repname+"d:"+first;
		    //if(latencies.containsKey(key)) {
		    //lat=((Double)latencies.get(key)).doubleValue();
		    //if (lat < max_latency && lat > min_latency) {
		    newList.add((String)oldList.elementAt(ii));
		    //}
		    //}
		}
	    }
	    
	    else {
		//if (DEBUG) System.err.println("this is bad!! "+key+" not found in latencies");
	    }
	}

	//Check to see if we have found enough nodes in candidate set
	if (newList.size() < max-count-1 && max!=999) {
	    //not enough candidates to continue
	    if(oldList.size()<max-count) {
		if(stack.size()>0 && count>0) {
		    newList.removeAllElements();
		    count--;
		    newVec.removeElementAt(newVec.size()-1);			  
		    oldList=(Vector)stack.elementAt(stack.size()-1);
		    stack.removeElementAt(stack.size()-1);
		    return recCheckLat(oldList, count, newVec, max, min_latency, max_latency, ignorelatency);
		}
		else {
		    return newVec;
		}
	    }
	    else {
		//this iteration failed, but we can try again
		newList.removeAllElements();
		return recCheckLat(oldList, count, newVec, max, min_latency, max_latency, ignorelatency);
	    }
	}	      
	else {
	    //so far so good...
	    newVec.add(first);
	    
	    Vector tempvec=new Vector();
	    for (int i=0; i<oldList.size(); i++) {
		tempvec.add(oldList.elementAt(i));
	    }
	    stack.add(tempvec);
	    count++;
	    
	    //success!
	    if(count==max || (newList.size()==0 && max==999)) {
		if(newList.size()==0) {
		    if(DEBUG) System.err.println("newlist size 0");
		}
		return newVec;		
	    }
	    else {
		return recCheckLat(newList, count, newVec, max, min_latency, max_latency, ignorelatency);
	    }
	}	      
    }
    
    public boolean checkConstraints(Vector oldSet, long stoptime, int totalnumgroups) {
	long curr_time=System.currentTimeMillis();
	Vector parentVec = new Vector();
	Vector nodesSeen = new Vector();
	Vector firstList[] = new Vector[10];
	Group firstGroup[] = new Group[10];
	int iter=0, level=0, max[]=new int[5];
	double matchcost, tempcost;

	for (int i=0; i<oldSet.size(); i++) {
	    firstList[i]=(Vector)oldSet.elementAt(i);
	    max[i]=firstList[i].size();
	    //max[i]=5;
	}

	for (int i=oldSet.size(); i<max.length; i++) {
	    max[i]=1;
	}

	if(oldSet.size()>0) {
	    for (int a=0; a<max[0]; a++) {
		if(DEBUG2) System.err.println("iter "+iter);
		iter++;
		level=1;
		parentVec.removeAllElements();
		firstGroup[level-1]=(Group)firstList[level-1].elementAt(a);
		nodesSeen=updateNodesSeen(firstGroup[level-1].getList(), parentVec);
		parentVec.add(firstGroup[level-1]);

		if(oldSet.size()>1) {
		    for (int b=0; b<max[1]; b++) {
			level=2;
			firstGroup[level-1]=(Group)firstList[level-1].elementAt(b);
			nodesSeen=updateNodesSeen(null,parentVec);
			if(checkHeuristics(nodesSeen,firstGroup[level-1].getList())) {
			    nodesSeen=updateNodesSeen(firstGroup[level-1].getList(), parentVec);
			    parentVec.add(firstGroup[level-1]);

			    if(oldSet.size()>2) {
				for (int c=0; c<max[2]; c++) {
				    level=3;
				    firstGroup[level-1]=(Group)firstList[level-1].elementAt(c);
				    nodesSeen=updateNodesSeen(null,parentVec);
				    if(checkHeuristics(nodesSeen,firstGroup[level-1].getList())) {
					nodesSeen=updateNodesSeen(firstGroup[level-1].getList(), parentVec);
					parentVec.add(firstGroup[level-1]);	

					if(oldSet.size()>3) {
					    for (int d=0; d<max[3]; d++) {
						level=4;					       
						firstGroup[level-1]=(Group)firstList[level-1].elementAt(d);
						nodesSeen=updateNodesSeen(null,parentVec);
						if(checkHeuristics(nodesSeen,firstGroup[level-1].getList())) {
						    nodesSeen=updateNodesSeen(firstGroup[level-1].getList(), parentVec);
						    parentVec.add(firstGroup[level-1]);	
						    
						    if(oldSet.size()>4) {
							for (int e=0; e<max[4]; e++) {
							    level=5;					       
							    firstGroup[level-1]=(Group)firstList[level-1].elementAt(e);
							    nodesSeen=updateNodesSeen(null,parentVec);
							    if(checkHeuristics(nodesSeen,firstGroup[level-1].getList())) {
								nodesSeen=updateNodesSeen(firstGroup[level-1].getList(), parentVec);
								parentVec.add(firstGroup[level-1]);	
						    
								if(oldSet.size()>5) {
								    for (int f=0; f<max[5]; f++) {
									level=6;					       
									firstGroup[level-1]=(Group)firstList[level-1].elementAt(f);
									nodesSeen=updateNodesSeen(null,parentVec);
									if(checkHeuristics(nodesSeen,firstGroup[level-1].getList())) {
									    nodesSeen=updateNodesSeen(firstGroup[level-1].getList(), parentVec);
									    parentVec.add(firstGroup[level-1]);	
									    
									    if(oldSet.size()>6) {
										System.err.println("TOO MANY GROUPS (for code to handle)!");
										System.exit(1);
									    }  
									    else if (parentVec.size()==groups.size()) {
										if(DEBUG) {
										    System.err.print("parent vec: ");
										    for (int s=0; s<parentVec.size(); s++) {
											System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
													 +((Group)parentVec.elementAt(s)).getID());
										    }
										    System.err.println();
										}
										matchcounter++;
										matchcost=checkLatencyConstraints(parentVec);
										if(matchcost!=-1) {	
										    tempcost=matchcost;
										    if (DEBUG2) System.err.print(matchcounter);
										    Vector tempvec = new Vector();
										    for (int s=0; s<parentVec.size(); s++) {
											tempcost+=((Group)parentVec.elementAt(s)).getCost();
											if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
														    +((Group)parentVec.elementAt(s)).getID());
											tempvec.add(((Group)parentVec.elementAt(s)).getID());
										    }
										    if(DEBUG2) System.err.print(" "+tempcost);
										    
										    if(tempcost<lowestcost) {
											lowestcost=tempcost;
											groupList.insertElementAt(tempvec,0);
											if(DEBUG2) System.err.println(" ***new lowest cost found "+((System.currentTimeMillis()-starttime)/1000.0));
										    }
										    else {
											groupList.add(tempvec);
											if(DEBUG2) System.err.println();
										    }
										    if(System.currentTimeMillis() > stoptime || groupList.size()==totalnumgroups) {
											if(DEBUG2) System.err.println("Out of time or Max groups found");
											return true;
										    }
										    if(DEBUG) System.err.println("added vec to grouplist "+parentVec);							  
										}
										else {
										    if(DEBUG) System.err.println("Skipping "+matchcounter+" due to latency");
										}							
										parentVec.remove(firstGroup[5]);
									    }	
									}
									else {
									    matchcounter+=1;
									    if(DEBUG) System.err.println("Skipping "+matchcounter+" "
													 +firstGroup[5].getName()+"-"+firstGroup[5].getID());
									}
									if(parentVec.size()==6){
									    parentVec.remove(parentVec.size()-1);
									}
								    }
								}
								else if (parentVec.size()==groups.size()) {
								    if(DEBUG) {
									System.err.print("parent vec: ");
									for (int s=0; s<parentVec.size(); s++) {
									    System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
											     +((Group)parentVec.elementAt(s)).getID());
									}
									System.err.println();
								    }
								    matchcounter++;
								    matchcost=checkLatencyConstraints(parentVec);
								    if(matchcost!=-1) {	
									tempcost=matchcost;
									if (DEBUG2) System.err.print(matchcounter);
									Vector tempvec = new Vector();
									for (int s=0; s<parentVec.size(); s++) {
									    tempcost+=((Group)parentVec.elementAt(s)).getCost();
									    if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
													+((Group)parentVec.elementAt(s)).getID());
									    tempvec.add(((Group)parentVec.elementAt(s)).getID());
									}
									if(DEBUG2) System.err.print(" "+tempcost);
									
									if(tempcost<lowestcost) {
									    lowestcost=tempcost;
									    groupList.insertElementAt(tempvec,0);
									    if(DEBUG2) System.err.println(" ***new lowest cost found "+((System.currentTimeMillis()-starttime)/1000.0));
									}
									else {
									    groupList.add(tempvec);
									    if(DEBUG2) System.err.println();
									}
									if(System.currentTimeMillis() > stoptime || groupList.size()==totalnumgroups) {
									    if(DEBUG2) System.err.println("Out of time or Max groups found");
									    return true;
									}
									if(DEBUG) System.err.println("added vec to grouplist "+parentVec);							  
								    }
								    else {
									if(DEBUG) System.err.println("Skipping "+matchcounter+" due to latency");
								    }							
								    parentVec.remove(firstGroup[4]);
								}	
							    }
							    else {
								matchcounter+=1;
								if(DEBUG) System.err.println("Skipping "+matchcounter+" "
											     +firstGroup[4].getName()+"-"+firstGroup[4].getID());
							    }
							    if(parentVec.size()==5){
								parentVec.remove(parentVec.size()-1);
							    }
							}
						    }
						    else if (parentVec.size()==groups.size()) {
							if(DEBUG) {
							    System.err.print("parent vec: ");
							    for (int s=0; s<parentVec.size(); s++) {
								System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
										 +((Group)parentVec.elementAt(s)).getID());
							    }
							    System.err.println();
							}
							matchcounter++;
							matchcost=checkLatencyConstraints(parentVec);
							if(matchcost!=-1) {	
							    tempcost=matchcost;
							    if (DEBUG2) System.err.print(matchcounter);
							    Vector tempvec = new Vector();
							    for (int s=0; s<parentVec.size(); s++) {
								tempcost+=((Group)parentVec.elementAt(s)).getCost();
								if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
											    +((Group)parentVec.elementAt(s)).getID());
								tempvec.add(((Group)parentVec.elementAt(s)).getID());
							    }
							    if(DEBUG2) System.err.print(" "+tempcost);
							    
							    if(tempcost<lowestcost) {
								lowestcost=tempcost;
								groupList.insertElementAt(tempvec,0);
								if(DEBUG2) System.err.println(" ***new lowest cost found "+((System.currentTimeMillis()-starttime)/1000.0));
							    }
							    else {
								groupList.add(tempvec);
								if(DEBUG2) System.err.println();
							    }
							    if(System.currentTimeMillis() > stoptime || groupList.size()==totalnumgroups) {
								if(DEBUG2) System.err.println("Out of time or Max groups found");
								return true;
							    }
							    if(DEBUG) System.err.println("added vec to grouplist "+parentVec);							  
							}
							else {
							    if(DEBUG) System.err.println("Skipping "+matchcounter+" due to latency");
							}							
							parentVec.remove(firstGroup[3]);
						    }	
						}
						else {
						    matchcounter+=1;
						    if(DEBUG) System.err.println("Skipping "+matchcounter+" "
										 +firstGroup[3].getName()+"-"+firstGroup[3].getID());
						}
						if(parentVec.size()==4){
						    parentVec.remove(parentVec.size()-1);
						}
					    }
											
					}
					else if (parentVec.size()==groups.size()) {
					    if(DEBUG) {
						System.err.print("parent vec: ");
						for (int s=0; s<parentVec.size(); s++) {
						    System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
								     +((Group)parentVec.elementAt(s)).getID());
						}
						System.err.println();
					    }
					    matchcounter++;
					    matchcost=checkLatencyConstraints(parentVec);
					    if(matchcost!=-1) {	
						tempcost=matchcost;
						if (DEBUG2) System.err.print(matchcounter);
						Vector tempvec = new Vector();
						for (int s=0; s<parentVec.size(); s++) {
						    tempcost+=((Group)parentVec.elementAt(s)).getCost();
						    if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
										+((Group)parentVec.elementAt(s)).getID());
						    tempvec.add(((Group)parentVec.elementAt(s)).getID());
						}
						if(DEBUG2) System.err.print(" "+tempcost);
						
						if(tempcost<lowestcost) {
						    lowestcost=tempcost;
						    groupList.insertElementAt(tempvec,0);
						    if(DEBUG2) System.err.println(" ***new lowest cost found "+((System.currentTimeMillis()-starttime)/1000.0));

						}
						else {
						    groupList.add(tempvec);
						    if(DEBUG2) System.err.println();
						}
						if(System.currentTimeMillis() > stoptime || groupList.size()==totalnumgroups) {
						    if(DEBUG2) System.err.println("Out of time or Max groups found");
						    return true;
						}
						if(DEBUG) System.err.println("added vec to grouplist "+parentVec);
					    }
					    else {
						if(DEBUG) System.err.println("Skipping "+matchcounter+" due to latency");
					    }					    
					    parentVec.remove(firstGroup[2]);
					}	
				    }
				    else {
					matchcounter+=max[3];
					if(DEBUG) System.err.println("Skipping "+matchcounter+" "
								     +firstGroup[2].getName()+"-"+firstGroup[2].getID());
				    }
				    if(parentVec.size()==3){
					parentVec.remove(parentVec.size()-1);
				    }
				}
			    }
			    else if (parentVec.size()==groups.size()) {
				if(DEBUG) {
				    System.err.print("parent vec: ");
				    for (int s=0; s<parentVec.size(); s++) {
					System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
							 +((Group)parentVec.elementAt(s)).getID());
				    }
				    System.err.println();
				}
				matchcounter++;
				matchcost=checkLatencyConstraints(parentVec);
				if(matchcost!=-1) {	
				    tempcost=matchcost;
				    if (DEBUG2) System.err.print(matchcounter);
				    Vector tempvec=new Vector();
				    for (int s=0; s<parentVec.size(); s++) {
					tempcost+=((Group)parentVec.elementAt(s)).getCost();
					if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
								    +((Group)parentVec.elementAt(s)).getID());
					tempvec.add(((Group)parentVec.elementAt(s)).getID());
				    }
				    if(DEBUG2) System.err.print(" "+tempcost);
				    
				    if(tempcost<lowestcost) {
					lowestcost=tempcost;
					groupList.insertElementAt(tempvec,0);
					if(DEBUG2) System.err.println(" ***new lowest cost found "+((System.currentTimeMillis()-starttime)/1000.0));
								
				    }
				    else {
					groupList.add(tempvec);
					if(DEBUG2) System.err.println();
				    }
				    if(System.currentTimeMillis() > stoptime || groupList.size()==totalnumgroups) {
					if(DEBUG2) System.err.println("Out of time or Max groups found");
					return true;
				    }
				    if(DEBUG) System.err.println("added vec to grouplist "+parentVec);							    
				}
				else {
				    if(DEBUG) System.err.println("Skipping "+matchcounter+" due to latency");
				}
				parentVec.remove(firstGroup[1]);
			    }				    
			}
			else {
			    matchcounter+=(max[2]*max[3]);
			    if(DEBUG) System.err.println("Skipping "+matchcounter+" "
							  +firstGroup[1].getName()+"-"+firstGroup[1].getID());
			}
			if(parentVec.size()==2){
			    parentVec.remove(parentVec.size()-1);
			}
		    }		    
		}
		else if (parentVec.size()==groups.size()) {
		    if(DEBUG) {
			System.err.print("parent vec: ");
			for (int s=0; s<parentVec.size(); s++) {
			    System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
					     +((Group)parentVec.elementAt(s)).getID());
			}
			System.err.println();
		    }
		    matchcounter++;
		    matchcost=checkLatencyConstraints(parentVec);
		    if(matchcost!=-1) {	
			tempcost=matchcost;
			if (DEBUG2) System.err.print(matchcounter);
			Vector tempvec = new Vector();
			for (int s=0; s<parentVec.size(); s++) {
			    tempcost+=((Group)parentVec.elementAt(s)).getCost();
			    if(DEBUG2) System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
							+((Group)parentVec.elementAt(s)).getID());
			    tempvec.add(((Group)parentVec.elementAt(s)).getID());
			}
			if(DEBUG2) System.err.print(" "+tempcost);
			
			if(tempcost<lowestcost) {
			    lowestcost=tempcost;
			    groupList.insertElementAt(tempvec,0);
			    if(DEBUG2) System.err.println(" ***new lowest cost found "+((System.currentTimeMillis()-starttime)/1000.0));
								
			}
			else {
			    groupList.add(tempvec);		       
			    if(DEBUG2) System.err.println();
			}
			if(System.currentTimeMillis() > stoptime || groupList.size()==totalnumgroups) {
			    if(DEBUG2) System.err.println("Out of time or Max groups found");
			    return true;
			}
			if(DEBUG) System.err.println("added vec to grouplist "+parentVec);							    
		    }
		    else {
			if(DEBUG) System.err.println("Skipping "+matchcounter+" due to latency");
		    }		   
		    parentVec.remove(firstGroup[0]);
		}
		else {
		    matchcounter+=(max[1]*max[2]*max[3]);
		    if(DEBUG) System.err.println("Skipping "+matchcounter+" "
						 +firstGroup[0].getName()+"-"+firstGroup[0].getID());
		}
		if(parentVec.size()==1){
		    parentVec.remove(parentVec.size()-1);
		}
	    }	     
	}
	else {
	    return false;
	}

	if(groupList.size()>0) {
	    if(DEBUG) System.err.println("grouplist size is "+groupList.size());
	    return true;
	}
	else {
	    return false;
	}
	
    }
    
    public Vector updateNodesSeen(Vector list, Vector parentVec) {
	Vector nodesSeen=new Vector();
	for (int i=0; i<parentVec.size(); i++) {
	    Group g = (Group)parentVec.elementAt(i);
	    Vector grlist = g.getList();
	    for (int j=0; j<grlist.size(); j++) {
		nodesSeen.add((String)grlist.elementAt(j));
	    }
	}
	if(list==null)
	    return nodesSeen;

	for (int i=0; i<list.size(); i++) {
	    nodesSeen.add((String)list.elementAt(i));
	}
	return nodesSeen;
    }
    
    public boolean checkConstraints(Vector oldSet, Vector parentVec) {		      
	//makes sure nodes don't appear in more than 1 group
	//also checks cross-group latency constraints

	if(DEBUG) {
	    System.err.print("start with stack "+constraintStack.size()+": ");
	    for (int s=0; s<parentVec.size(); s++) {
		System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
				 +((Group)parentVec.elementAt(s)).getID());
	    }
	    System.out.println();
	}

	Vector nodesSeen = new Vector();
	for (int q=0; q<parentVec.size(); q++) {
	    Group tempgroup = (Group)parentVec.elementAt(q);
	    Vector nodeList = tempgroup.getList();
	    for (int qq=0; qq<nodeList.size(); qq++) {
		nodesSeen.add((String)nodeList.elementAt(qq));
	    }
	}
	
	//Pop first vector for first group off vector of potential nodes 
	Vector firstList=(Vector)oldSet.remove(0);		

	if(oldSet.size()>0) {
	    Group firstGroup = null;
	    Vector firstGroupList = null;

	    if(firstList.size()>0) {
		firstGroup = (Group)firstList.remove(0);
		firstGroupList = firstGroup.getList();
		
		if (!checkHeuristics(firstGroupList, nodesSeen) && firstList.size()==0) {
		    if(DEBUG) System.err.println("skipping "+firstGroup.getName()+"-"+firstGroup.getID());		
		}

		while (!checkHeuristics(firstGroupList, nodesSeen) && firstList.size()>0) {
		    if(DEBUG) System.err.println("skipping "+firstGroup.getName()+"-"+firstGroup.getID());		
		    firstGroup = (Group)firstList.remove(0);
		    firstGroupList = firstGroup.getList();
		}

		if(!checkHeuristics(firstGroupList, nodesSeen)) {
		    if(DEBUG) System.err.println("skipping "+firstGroup.getName()+"-"+firstGroup.getID());		
		    firstGroup=null;
		}
	    }
	    
	    if(firstGroup==null && constraintStack.size() > 0) {
		if(DEBUG) System.err.println("here at the bottom and stack "+constraintStack.size());

		Vector stackHash = (Vector)constraintStack.pop();
		
		if(DEBUG) {
		    System.err.print("popping ");
		    
		    for (int i=0; i<stackHash.size(); i++) {
			Vector temp = (Vector)stackHash.elementAt(i);
			String name="null";
			if(temp.size()>0) {
			    Group tempG = (Group)temp.elementAt(0);
			    name=tempG.getName();
			}
			if(DEBUG) System.err.print(name+" "+temp.size()+" ");
		    }
		    if(DEBUG) System.err.println();
		}
		    
		while(stackHash.size()+parentVec.size() > groups.size()) {
		    parentVec.remove(parentVec.size()-1);
		}
		
		return checkConstraints(stackHash, parentVec);
	    }
	    
	    else {
		if(checkHeuristics(firstGroupList, nodesSeen)) {
		    if(DEBUG) System.err.println("continuing with "+firstGroup.getName()+"-"+firstGroup.getID());	   	    
		    parentVec.add(firstGroup);
		    
		    Vector stackHash = new Vector();
		    Vector tempvec = new Vector();
		    for (int v=0; v<firstList.size(); v++) {
			tempvec.add(firstList.elementAt(v));
		    }
		    stackHash.add(tempvec);
		    if(DEBUG) System.err.print("pushing "+firstGroup.getName()+" "+tempvec.size()+" ");
		    for (int i=0; i<oldSet.size(); i++) {
			tempvec = (Vector) oldSet.elementAt(i);
			Group tempg = (Group)tempvec.elementAt(0);
			Vector newvec = new Vector();
			for (int v=0; v<tempvec.size(); v++) {
			    newvec.add(tempvec.elementAt(v));
			}
			stackHash.add(newvec);
			if(DEBUG) System.err.print(tempg.getName()+" "+newvec.size()+" ");
		    }
		    if(DEBUG) System.err.println();
		    
		    constraintStack.push(stackHash);
		    return checkConstraints(oldSet, parentVec);
		}
		else {
		    if(DEBUG) System.err.println("WE ARE DONE");
		}		
	    }
	}
	
	else if (parentVec.size()==groups.size()-1) {
	    Group firstGroup=null;
	    Vector firstGroupList=null;
	    for (int i=0; i<firstList.size(); i++) {
		firstGroup = (Group)firstList.elementAt(i);
		firstGroupList = firstGroup.getList();
		
		if(checkHeuristics(firstGroupList, nodesSeen)) {
		    parentVec.add(firstGroup);
		
		    if(DEBUG) {
			System.err.print("parent vec: ");
			for (int s=0; s<parentVec.size(); s++) {
			    System.err.print(" "+((Group)parentVec.elementAt(s)).getName()+"-"
					     +((Group)parentVec.elementAt(s)).getID());
			}
			System.err.println();
		    }
		    
		    if(parentVec.size()==groups.size()){
			Vector v = new Vector();
			
			for (int s=0; s<parentVec.size(); s++) {
			     v.add(((Group)parentVec.elementAt(s)).getID());
			}
			groupList.add(v);
			parentVec.remove(firstGroup);
		    }
		}	
		 
		else {
		    if(DEBUG) System.err.println("skipping "+firstGroup.getName()+"-"+firstGroup.getID());		
		}
	    }	     
	
	    if(constraintStack.size()>0) {
		Vector stackHash = (Vector)constraintStack.pop();
		
		if(DEBUG) {
		    System.err.print("popping ");
		    
		    for (int i=0; i<stackHash.size(); i++) {
			Vector temp = (Vector)stackHash.elementAt(i);
			String name="null";
			if(temp.size()>0) {
			    Group tempG = (Group)temp.elementAt(0);
			    name=tempG.getName();
			}
			if(DEBUG) System.err.print(name+" "+temp.size()+" ");
		    }
		    if(DEBUG) System.err.println();
		}
		
		while(stackHash.size()+parentVec.size() > groups.size()) {
		    parentVec.remove(parentVec.size()-1);
		}
		
		return checkConstraints(stackHash, parentVec);
	    }
	}	
	else {
	    if (DEBUG) System.err.println("ERROR HERE!");
	}

	DEBUG=false;
	if(groupList.size()>0) {
	    return true;
	}
	else 
	    return false;
    }
    
    public boolean checkConstraintsOld(HashMap oldSet, Vector parentVec, Stack stack) {		      
	//makes sure nodes don't appear in more than 1 group
	//also checks cross-group latency constraints
       	
	HashMap newSet=new HashMap();
	Iterator it=oldSet.keySet().iterator();
	boolean done=false;
	String secondname="";
	Vector secondlist=null;
	Vector nodesSeen = new Vector();

	//Pop first element off list of potential nodes 
	Vector firstlist=(Vector)oldSet.remove((String)it.next());	
	
	it=oldSet.keySet().iterator();

	if(!it.hasNext()) {
	    done=true;
	}
	
	if(!done) {
	    secondname=(String)it.next();
	    secondlist=(Vector)oldSet.get(secondname);

	    if(firstlist.size()>0) {
		HashMap savedSet = new HashMap();
		Group removeFirst = (Group)firstlist.remove(0);
		Vector newfirstlist = new Vector();
		for (int t=0; t<firstlist.size(); t++) {
		    newfirstlist.add(firstlist.elementAt(t));
		}
		savedSet.put(removeFirst.getName(), newfirstlist);	
		savedSet.put(secondname, secondlist);
		while(it.hasNext()) {
		    String tempname = (String)it.next();
		    savedSet.put(tempname, oldSet.get(tempname));
		}

		if(DEBUG)System.err.println("removing "+removeFirst.getName()+" "+removeFirst.getID());
		stack.push(savedSet);
		firstlist.insertElementAt(removeFirst, 0);

		it=oldSet.keySet().iterator();
		secondname=(String)it.next();
		secondlist=(Vector)oldSet.get(secondname);
	    }
	}

	for(int i=0; i<firstlist.size(); i++) {
	    Vector newSetVec=new Vector();
	    Group first=(Group)firstlist.elementAt(i);	    
	    if(DEBUG) System.err.println("looking at group "+first.getName()+"-"+first.getID()+" i="+i);
	    parentVec.add(first);

	    nodesSeen.removeAllElements();
	    for (int q=0; q<parentVec.size(); q++) {
		Group tempgroup = (Group)parentVec.elementAt(q);
		Vector nodeList = tempgroup.getList();
		for (int qq=0; qq<nodeList.size(); qq++) {
		    nodesSeen.add((String)nodeList.elementAt(qq));
		}
	    }

	    if(!done) {
		for(int j=0; j<secondlist.size(); j++) {	
		    Group second=(Group)secondlist.elementAt(j);
		    Vector nodelist2=second.getList();
		    boolean ret=checkHeuristics(nodelist2, nodesSeen);
		    if(ret) {
			newSetVec.add(second);
		    }
		    else {
			if(DEBUG) System.err.println("removing "+second.getName()+"-"+second.getID());
		    }
		}
		newSet.put(secondname,newSetVec);

		//recurse and check for the rest of the groups
		while(it.hasNext()) {
		    String name=(String)it.next();
		    Vector v=(Vector)oldSet.get(name);
		    newSet.put(name,v);
		}
		return checkConstraintsOld(newSet, parentVec, stack);
	    }

	    if(done && parentVec.size()==groups.size()) {
		//now check cross-group latency
		matchcounter++;
		//		List xlist = doc.getRootElement().getChildren("constraint");
		List xlist = constraintList;
		double matchcost = checkLatencyConstraints(parentVec);
		if(xlist == null || xlist.size() == 0 || matchcost != -1) {
		    double tempcost=matchcost;
		    Vector v = new Vector();
		    if (DEBUG) System.err.print(matchcounter);
		    for (int s=0; s<parentVec.size(); s++) {
			tempcost+=((Group)parentVec.elementAt(s)).getCost();
			v.add(((Group)parentVec.elementAt(s)).getID());
			if (DEBUG) System.err.print(" "+((Group)v.elementAt(s)).getName()+"-"
					 +((Group)v.elementAt(s)).getID());
		    }
		    if (DEBUG) System.err.println(" "+tempcost);

		    //enable this is we want to stop after first solution
		    //return true;
		    if(tempcost<lowestcost) {
			lowestcost=tempcost;
			groupList.insertElementAt(v,0);
			// 	System.err.print("new lowest match "+lowestcost);
			// 			for (int j=0; j<v.size(); j++) {
			// 			    System.err.print(" "+((Group)v.elementAt(j)).getName()+"-"
			// 					       +((Group)v.elementAt(j)).getID());
			
			// 			}
			// 			System.err.println();
		    }
		    else {
			groupList.add(v);
		    }
		}
		else {
		    //System.err.println(matchcounter+" "+matchcost);
		}
	    }	 
	    parentVec.remove(first);
	}
	if(stack.size()>0) {
	    HashMap lastSet = (HashMap)stack.pop();
	    return checkConstraintsOld(lastSet, new Vector(), stack);
	}
	
	else {
	    if(groupList.size()>0) {
		return true;
	    }
	    else 
		return false;
	}
    }

    public boolean checkSite(Vector v, int numhosts) {
	sites=new HashMap();
	
	for (int i=0; i<v.size(); i++) {
	    Vector list = ((Group)v.elementAt(i)).getList();
	    for (int j=0; j<list.size(); j++) {		
		HashMap h1=(HashMap)replyStats.get((String)list.elementAt(j));
		Double svsite = (Double)h1.get("site");
		double site = svsite.doubleValue();
		Integer count=((Integer)sites.get(new Double(site))); 
		int ccount=0;
		
		//check numhosts
		if(count==null) {
		    sites.put(new Double(site),new Integer(1));
		}
		
		else { 		
		    ccount=count.intValue();
		    ccount++;
		    count=new Integer(ccount);
		    sites.put(new Double(site), count);
		}
		
		//if we have seen too many from this site, remove node from candidate set
		//otherwise, add to v2
		if(ccount<=numhosts) {
		    //continue
		}
		else {
		    //System.err.println((String)list.elementAt(j));
		    return false;
		}
	    }
	}
	sites=null;
	return true;
    }
    
    public boolean checkHeuristics(Vector v1, Vector v2) {
	if(v1==null || v2==null) {
	    return false;
	}

	for (int i=0; i<v1.size(); i++) {
	    if(v2.contains(v1.elementAt(i))){
		if(DEBUG) System.err.print("Repeat element "+v1.elementAt(i)+" ");
		return false;
	    }
	}
	return true;
    }

    public double checkLatencyConstraints(Vector grouplist) {	
	if(grouplist.size()<=1) {
	    return 0;
	}
	
	Vector groupNames=new Vector();
	Vector results=new Vector();
	double abs_min_latency=0.0, abs_max_latency=0.0, ideal_min_latency=0.0, ideal_max_latency=0.0, lat_cost=0.0;	
	boolean stop=false, ignorelatency=false;
	List list=constraintList;
	Iterator it=list.iterator();
	HashMap glist=new HashMap();
	
	for (int i=0; i<grouplist.size(); i++) {
	    String s=((Group)grouplist.elementAt(i)).getName();
	    String s2=String.valueOf(((Group)grouplist.elementAt(i)).getID());
	    //System.err.println(s+"-"+s2);
	    glist.put(s,(Group)grouplist.elementAt(i));
	}

	while(it.hasNext() && !stop) {
	    Element ele=(Element)it.next();
	    String gn = ele.getChildTextTrim("group_names");
	    String pairs="one";
	    Element pair = ele.getChild("pairs");
	    if(pair!=null)	    
		pairs = ele.getChildTextTrim("pairs");
	    Group g1=null, g2=null;
	    groupNames.removeAllElements();
	
	    StringTokenizer strtok=new StringTokenizer(gn," ");	    
	    while(strtok.hasMoreTokens()) {
		String s = strtok.nextToken();
		
		if(!groupNames.contains(s)) {
		    groupNames.add(s);
		}	    	    
		else {
		    result+="Unable to complete request.  No group information for "+s+"\n";
		    return -1;
		}
	    }

	    String s = ele.getChildTextTrim("numhosts");
	    if (s!=null) {
		int numhosts = Integer.parseInt(s);
		if (DEBUG) System.err.println("numhosts="+numhosts);	  
		if(!checkSite(grouplist, numhosts)) {
		    if(DEBUG2) {
			System.err.print("Too many hosts "+matchcounter);
			for (int j=0; j<grouplist.size(); j++) {
			    System.err.print("  "+((Group)grouplist.elementAt(j)).getName()+"-"
					     +((Group)grouplist.elementAt(j)).getID());
			}
			System.err.println();
		    }
		    return -1;
		}
	    }
	    else {
		//System.err.println("numhosts not specified");
	    }

	    //Now we need to parse the latency requirement 
	    String l = ele.getChildTextTrim("latency");
	    if(l==null) {
		l="0.0,0.0,999999.0,999999.0,0.0";
		ignorelatency=true;
	    }

	    if(ignorelatency)
		return 0;

	    StringTokenizer st= new StringTokenizer(l, ",");
	    abs_min_latency=Double.valueOf(st.nextToken()).doubleValue();
	    ideal_min_latency=Double.valueOf(st.nextToken()).doubleValue();
	    ideal_max_latency=Double.valueOf(st.nextToken()).doubleValue();
	    abs_max_latency=Double.valueOf(st.nextToken()).doubleValue();
	    lat_cost=Double.valueOf(st.nextToken()).doubleValue();
	    
	    //Now see if links exist between pairs of groups that meet latency requirement
	    int c=0, d=0;
	    for(int n=0; n<groupNames.size()-1; n++) {
		for (int i=n+1; i<groupNames.size(); i++) {  
		    Vector ret=checkGroupPair((Group)glist.get((String)groupNames.elementAt(n)), 
					      (Group)glist.get((String)groupNames.elementAt(i)),
					      abs_min_latency, ideal_min_latency, ideal_max_latency, abs_max_latency, lat_cost, pairs); 
		    if(ret!=null) {
			c++;
			results.add(ret);
		    }
		    else {
			if (DEBUG) System.err.println("latency failed");
			return -1;
		    }
		    d++;
		}
	    }
	    if(c!=d) {
		stop=true;
	    }
	}
	double cost=0, gcost=0;
	if(!stop) {
	    for (int i=0; i<results.size(); i++) {
		Vector temp=(Vector)results.elementAt(i);
		gcost=((Double)temp.elementAt(0)).doubleValue();
		cost+=gcost;

		String temp_result="";
		// temp_result+="\nLink that meets min/max latency requirement of "
		// 		    +abs_min_latency+"/"+abs_max_latency+" ms:\n";
		// 		temp_result+="Between "+temp.elementAt(1);
		// 		temp_result+=" and "+temp.elementAt(2)+"\n";
		// 		temp_result+=temp.elementAt(1)+" member:  "+temp.elementAt(3);
		// 		temp_result+="\n"+temp.elementAt(2)+" member:  "+temp.elementAt(4);
		// 		temp_result+="\nAvg Latency:  "+temp.elementAt(5);
		// 		temp_result+="\nLink Cost: "+gcost;
		// 		temp_result+="\n"+(((Double)temp.elementAt(10)).doubleValue())
		// 		    +" Total cost: "+(((Double)temp.elementAt(8)).doubleValue()
		// 				      +((Double)temp.elementAt(9)).doubleValue()+gcost)+
		// 		    " Iteration number "+((Double)temp.elementAt(11)).doubleValue();
		// 		temp_result+="\nGroup match #"+(((Double)temp.elementAt(10)).doubleValue())+"\n";
		
		temp_result=temp.elementAt(1)+" "+temp.elementAt(2)+" "
		    +((Double)temp.elementAt(8)).doubleValue()+" "+((Double)temp.elementAt(9)).doubleValue()+" "+gcost+" "
		    +temp.elementAt(10)+" "+temp.elementAt(11);
		
		if(DEBUG) System.err.println(temp_result);
		//result+=temp_result;
	    }
	    return cost;
	}
	return -1;
    }
    
    public Vector checkGroupPair(Group g1, Group g2, double abs_min_latency, 
			      double ideal_min_latency, double ideal_max_latency, 
				 double abs_max_latency, double lat_cost, String pairs) {	
	Vector group1=g1.getList();
	Vector group2=g2.getList();
	Vector returnVec=new Vector();
	groupcounter2++;

	for (int j=0; j<group2.size(); j++) {
	    if(group1.contains(group2.elementAt(j))) {
		return null;
	    }	    
	}

	double lowest_cost=99999;
	for (int j=0; j<group1.size(); j++){
	    for (int k=j+1; k<group2.size(); k++) {

		double latency=-1.0, rev_latency=-1.0, avg_latency=0;
		if(latencies.containsKey("s:"+(String)group1.elementAt(j)+"d:"+(String)group2.elementAt(k))) {
		    latency=((Double)latencies.get("s:"+(String)group1.elementAt(j)+"d:"
						   +(String)group2.elementAt(k))).doubleValue();
		}
		else if(latencies.containsKey("s:"+(String)group2.elementAt(k)+"d:"+(String)group1.elementAt(j))) {
		    rev_latency=((Double)latencies.get("s:"+(String)group2.elementAt(k)+"d:"
						       +(String)group1.elementAt(j))).doubleValue();
		}
		
		double cost=0.0;
		if(latency > -1) {
		    avg_latency = latency;
		}
		if(rev_latency > -1) {
		    avg_latency = rev_latency;
		}
		if (latency > -1 && rev_latency > -1) {
		    avg_latency = (latency+rev_latency)/2.0;
		}

		double denom=0;
		if(avg_latency<abs_max_latency && avg_latency > abs_min_latency) {
		    if(avg_latency<ideal_min_latency) {
			denom = ideal_min_latency-abs_min_latency;
			cost=((ideal_min_latency-avg_latency)/denom)*lat_cost;  
		    }
		    if(avg_latency>ideal_max_latency) {
			denom=abs_max_latency-ideal_max_latency;
			cost=((avg_latency-ideal_max_latency)/denom)*lat_cost;
		    }
		    
		    if(cost<lowest_cost) {
			returnVec.removeAllElements();
			returnVec.add(new Double(cost));
			returnVec.add(g1.getName()+"-"+g1.getID());
			returnVec.add(g2.getName()+"-"+g2.getID());
			returnVec.add(group1.elementAt(j));
			returnVec.add(group2.elementAt(k));
			//			System.out.println(avg_latency);
			//			returnVec.add(String.valueOf(avg_latency));
			returnVec.add("lat");
			returnVec.add(g1.getList());
			returnVec.add(g2.getList());
			returnVec.add(new Double(g1.getCost()));
			returnVec.add(new Double(g2.getCost()));
		    }
		}
		else {
		    if(pairs.equals("all")) {
			if(DEBUG)System.err.println("all latency pairs failed for "+g1.getName()+"-"+g1.getID()+" and "+g2.getName()+"-"+g2.getID());
			return null;
		    }
		}
	    }
	}
	if(returnVec.size()>0) {
	    groupcounter++;
	    returnVec.add(new Double(groupcounter));
	    returnVec.add(new Double(groupcounter2));
	    return returnVec;
	}
	
	else 
	    return null;
    }
    
    public String getResult() {
	return result;
    }      

    public HashMap getMasterGrouphash() {
	return masterGrouphash;
    }

    public double getTotalCost() {
	return lowestcost;
    }
}
 
class Link {
    String src;
    String dst;
    double latency;

    public Link(String src, String dst, double latency) {
	this.src=src;
	this.dst=dst;
	this.latency=latency;
    }

    public double getLatency() {
	return latency;
    }

    public String getDst() {
	return dst;
    }

    public String getSrc() {
	return src;
    }
}

class Node {
    String name;
    double cost;

    public Node(String name, double cost) {
	this.name=name;
	this.cost=cost;
    }

    public double getCost() {
	return cost;
    }
    
    public void setCost(double c) {
	this.cost=c;
    }

    public String getName() {
	return name;
    }

    public String toString() {
        return "(Node  name="+name+"  cost="+cost+")";
    }    
}

//got this from the web
class QuickSort {
    static String type="";

    public QuickSort(String type) {
	this.type=type;
    }
    
    public Vector sort(Vector array){
	sort(array, 0, array.size() - 1);
	return array;
    }
    
    public static void sort(Vector array, int start, int end){
	int p;
	if (end > start){
	    p = partition(array, start, end);
	    sort(array, start, p-1);
	    sort(array, p+1, end);
	}
    }
    
    protected static int compare(double a, double b) {
	Double aa=new Double(a);
	Double bb=new Double(b);
	return aa.compareTo(bb);
    }
    
    protected static int partition(Vector array, int start, int end) {
	int left=0, right=0;
	double partitionElement=0.0;
	
	if(type.equals("Node")) {
	    // Arbitrary partition start...there are better ways...
	    partitionElement = ((Node)array.elementAt(end)).getCost();
	    
	    left = start - 1;
	    right = end;
	    for (;;) {
		while (compare(partitionElement, ((Node)array.elementAt(++left)).getCost()) == 1) {
		    if (left == end) break;
		}
		while (compare(partitionElement, ((Node)array.elementAt(--right)).getCost()) == -1){
		    if (right == start) break;
		}
		if (left >= right) break;
		swap(array, left, right);
	    }
	}
	else if (type.equals("Link")){
	    partitionElement = ((Link)array.elementAt(end)).getLatency();
	    
	    left = start - 1;
	    right = end;
	    for (;;) {
		while (compare(partitionElement, ((Link)array.elementAt(++left)).getLatency()) == 1) {
		    if (left == end) break;
		}
		while (compare(partitionElement, ((Link)array.elementAt(--right)).getLatency()) == -1){
		    if (right == start) break;
		}
		if (left >= right) break;
		swap(array, left, right);
	    }
	}
	else if (type.equals("Group")){
	    partitionElement = ((Group)array.elementAt(end)).getCost();
	    
	    left = start - 1;
	    right = end;
	    for (;;) {
		while (compare(partitionElement, ((Group)array.elementAt(++left)).getCost()) == 1) {
		    if (left == end) break;
		}
		while (compare(partitionElement, ((Group)array.elementAt(--right)).getCost()) == -1){
		    if (right == start) break;
		}
		if (left >= right) break;
		swap(array, left, right);
	    }
	}
	swap(array, left, end);
	    
	return left;
    }
    
    protected static void swap(Vector array, int i, int j) {
	Object temp;	
	temp = array.elementAt(i);
	array.setElementAt(array.elementAt(j), i);
	array.setElementAt(temp, j);
    }
}

